import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:osam/util/comparable_wrapper.dart';
import 'package:osam/util/state_stream.dart';

typedef ValueMapper<ST, V> = V Function(ST state);

// ignore: must_be_immutable
abstract class BaseState<ST> extends Equatable {
  int _lastKnownHashCode = -1;

  // ignore: close_sinks
  final _stateBroadcaster = StreamController<ST>.broadcast();

  @override
  List<Object> get props;

  StateStream<ST> get stateStream =>
      StateStream(_stateBroadcaster.stream.map((state) => state), this as ST);

  StateStream<V> propertyStream<V>(ValueMapper<ST, V> mapper) {
    final firstValue = mapper(stateStream.initialData);
    return StateStream(
        (PropertyStream<V>(
                stateStream.stream.map<V>((state) {
                  try {
                    return mapper(state);
                  } catch (e) {
                    handleError(e);
                    return null;
                  }
                }).where((property) => property != null),
                firstValue))
            .propertyStream,
        firstValue);
  }

  // mutable props workaround for hive adaptor
  void refreshHashcode() => _lastKnownHashCode = hashCode;

  void update() {
    if (_lastKnownHashCode != hashCode) {
      _stateBroadcaster.sink.add(this as ST);
      _lastKnownHashCode = hashCode;
    }
  }

  void handleError(Object error) => print(error);

  void destruct() => _stateBroadcaster.close();

  void clear() {}
}

class PropertyStream<V> {
  final Stream<V> inputStream;
  final V firstValue;
  int _valueHashCode = -1;

  PropertyStream(this.inputStream, this.firstValue) {
    _valueHashCode = ComparableWrapper(firstValue).hashCode;
  }

  Stream<V> get propertyStream => inputStream.map((next) {
        final comparableWrapper = ComparableWrapper(next);
        final result = _valueHashCode == comparableWrapper.hashCode;
        _valueHashCode = comparableWrapper.hashCode;
        return result ? null : next;
      }).where((event) => event != null);
}
